"""
Dummy metadata store module.
"""
from __future__ import annotations

from nefertem_core.stores.output.objects._base import OutputStore


class DummyOutputStore(OutputStore):
    """
    Dummy metadata store object implementation.

    Only allows the client to interact with store methods.
    """

    def __init__(self, path: str) -> None:
        super().__init__(path)

    def init_run(self, *args) -> None:
        """
        Placeholder method.

        Returns
        -------
        None
        """

    def log_metadata(self, *args) -> None:
        """
        Placeholder method.

        Returns
        -------
        None
        """

    def persist_artifact(self, *args) -> None:
        """
        Placeholder method.

        Returns
        -------
        None
        """



"""
Client module.
"""
from __future__ import annotations

import typing

from nefertem_core.run.builder import run_builder
from nefertem_core.stores.builder import store_builder

if typing.TYPE_CHECKING:
    from nefertem_core.run.run import Run


class Client:
    """
    The Client is a public interface that exposes methods to create runs and allows
    the user to add input stores to the pool of runs stores.

    Parameters
    ----------
    path : str
        Path where to store metadata and artifacts.
    stores : list[dict]
        List of dict containing configuration for the input stores.

    Methods
    -------
    add_store
        Add a new store to the client internal registry.
    create_run
        Create a new run.
    """

    def __init__(
        self,
        path: str | None = None,
        stores: list[dict] | None = None,
    ) -> None:
        self._tmp_dir = "./ntruns/tmp"
        self._setup_stores(path, stores)

    def _setup_stores(self, path: str | None = None, configs: list[dict] | None = None) -> None:
        """
        Build stores according to configurations provided by user and register
        them into the store registry.

        Parameters
        ----------
        path : str
            Path where to store metadata and artifacts.
        configs : list[dict]
            List of dict containing configuration for the input stores.
        """

        # Build output store
        store_builder.build_output_store(path)

        # Build input stores
        try:
            for cfg in configs:
                store_builder.build_input_store(self._tmp_dir, cfg)
        except TypeError:
            pass

    def add_store(self, config: dict) -> None:
        """
        Add a new store to the store registry.

        Parameters
        ----------
        config : dict
            Dictionary containing the configuration of the store.

        Returns
        -------
        None
        """
        store_builder.build_input_store(self._tmp_dir, config)

    def create_run(
        self,
        resources: list[dict],
        run_config: dict,
        experiment: str | None = None,
        run_id: str | None = None,
        overwrite: bool = False,
    ) -> Run:
        """
        Create a new run.

        Parameters
        ----------
        resources : list[dict]
            List of DataResource objects.
        run_config : dict
            RunConfig object.
        experiment : str
            Name of the experiment. An experiment is a logical unit for ordering the runs execution,
            by default "experiment".
        run_id : str
            Optional string parameter for user defined run id.
        overwrite : bool
            If True, the run metadata/artifact can be overwritten by a run with the same id.

        Returns
        -------
        Run
            Run object.
        """
        return run_builder.create_run(resources, run_config, self._tmp_dir, experiment, run_id, overwrite)



from __future__ import annotations

from nefertem_core.client.client import Client


def create_client(
    output_path: str | None = None,
    stores: list[dict] | None = None,
) -> Client:
    """
    Create a new Client object.

    Parameters
    ----------
    output_path : str
        Path to the metadata store.
    stores : list[dict]
        List of dict containing configuration for the artifact stores.

    Returns
    -------
    Client
        Client object.
    """
    return Client(output_path, stores)



"""
EnvLog module.
"""
from __future__ import annotations

import os
import platform

from psutil import virtual_memory


class Env:
    """
    Basic log structure for execution enviroment.

    Attributes
    ----------
    platform : str
        Execution platform.
    python_version : str
        Python version.
    cpu_model : str
        CPU model.
    cpu_core : int
        Number of CPU cores.
    ram : str
        RAM memory in GB.
    """

    def __init__(self) -> None:
        """
        Constructor.
        """
        self.platform = platform.platform()
        self.python_version = platform.python_version()
        self.cpu_model = platform.processor()
        self.cpu_core = os.cpu_count()
        self.ram = self.round_ram()

    @staticmethod
    def round_ram() -> str:
        """
        Return rounded GB ram memory.

        Returns
        -------
        str
            Rounded GB ram memory.
        """
        mem = virtual_memory().total / (1024.0**3)
        return f"{mem:.1f} GB"

    def to_dict(self) -> dict:
        return self.__dict__



"""
Run handler module.
"""
from __future__ import annotations

import concurrent.futures
import typing
from typing import Any

from nefertem_core.plugins.factory import builder_factory
from nefertem_core.plugins.utils import ResultType
from nefertem_core.stores.builder import get_all_input_stores
from nefertem_core.utils.utils import flatten_list, listify

if typing.TYPE_CHECKING:
    from nefertem_core.plugins.builder import PluginBuilder
    from nefertem_core.plugins.plugin import Plugin
    from nefertem_core.run.config import RunConfig


class RunHandler:
    """
    Run handler.

    This class create a layer of abstraction between the Run and its plugins.

    Attributes
    ----------
    _config : RunConfig
        Run configuration.
    _tmp_dir : str
        Temporary directory to store artifacts.
    _registry : dict
        Resul registry.
    """

    def __init__(self, config: RunConfig) -> None:
        """
        Constructor.
        """
        self._config = config
        self._registry = {}

    #############################
    # Execution methods
    #############################

    def run(self, *args, **kwargs) -> None:
        """
        Run plugins.

        Args and kwargs are passed to plugins, so if you use a plugin that require
        resources, you can pass them as arguments or keyword arguments.

        E.g.:
        ``
        run_handler.run(resources)
        run_handler.run(resources=resources)
        ``

        Parameters
        ----------
        args : Any
            Arguments.
        kwargs : Any
            Keyword arguments.
        """
        builders = self._get_builder()
        plugins = self._create_plugins(builders, *args, **kwargs)
        self._scheduler(plugins)

    def _get_builder(self) -> list[PluginBuilder]:
        """
        Return a list of builders.

        Returns
        -------
        list[PluginBuilder]
            List of builders.
        """
        return builder_factory(self._config, get_all_input_stores())

    @staticmethod
    def _create_plugins(builders: PluginBuilder, *args, **kwargs) -> list[Plugin]:
        """
        Return a list of plugins.
        """
        return flatten_list([builder.build(*args, **kwargs) for builder in builders])

    def _scheduler(self, plugins: list[Plugin]) -> None:
        """
        Schedule execution to avoid multiprocessing issues.
        """
        multiprocess = []
        multithreading = []
        sequential = []

        for plugin in plugins:
            # Concurrent execution
            if self._config.parallel:
                if plugin.exec_multiprocess:
                    multiprocess.append(plugin)
                elif plugin.exec_multithread:
                    multithreading.append(plugin)
            # Sequential execution
            else:
                sequential.append(plugin)

        self._sequential_execute(sequential)
        self._pool_execute_multithread(multithreading)
        self._pool_execute_multiprocess(multiprocess)

    def _sequential_execute(self, plugins: list[Plugin]) -> None:
        """
        Execute operations in sequence.
        """
        for plugin in plugins:
            data = self._execute(plugin)
            self._register_results(data)

    def _pool_execute_multiprocess(self, plugins: list[Plugin]) -> None:
        """
        Instantiate a concurrent.future.ProcessPoolExecutor pool to execute operations in
        multiprocessing.
        """
        with concurrent.futures.ProcessPoolExecutor(max_workers=self._config.num_worker) as pool:
            for data in pool.map(self._execute, plugins):
                self._register_results(data)

    def _pool_execute_multithread(self, plugins: list[Plugin]) -> None:
        """
        Instantiate a concurrent.future.ThreadPoolExecutor pool to execute operations in
        multithreading.
        """
        with concurrent.futures.ThreadPoolExecutor(max_workers=self._config.num_worker) as pool:
            for data in pool.map(self._execute, plugins):
                self._register_results(data)

    @staticmethod
    def _execute(plugin: Plugin) -> dict:
        """
        Wrap plugins main execution method. The handler create builders to build plugins.
        Once the plugin are built, the handler execute the main plugin operation, produce
        a nefertem report, render the execution artifact ready to be stored and save some
        library infos.
        """
        return plugin.execute()

    #############################
    # Registry methods
    #############################

    def _register_results(self, result: dict) -> None:
        """
        Register results.

        Parameters
        ----------
        result : dict
            Result dictionary.

        Returns
        -------
        None
        """
        for key, value in result.items():
            if key not in self._registry:
                self._registry[key] = []
            if isinstance(value, list):
                self._registry[key].extend(value)
            else:
                self._registry[key].append(value)

    def get_item(self, item_type: str) -> list[Any]:
        """
        Get item from registry.

        Parameters
        ----------
        item_type : str
            Item type.

        Returns
        -------
        list[Any]
            List of object.
        """
        objects = self._registry.get(item_type, [])

        # Get artifacts and nefertem report
        if item_type in [ResultType.FRAMEWORK.value, ResultType.NEFERTEM.value]:
            return [obj.artifact for obj in objects]

        # Get rendered
        return listify(flatten_list([obj.artifact for obj in objects]))

    def get_libraries(self) -> list[dict]:
        """
        Get libraries used by operations.

        Returns
        -------
        list[dict]
            List of libraries.
        """
        libs = []
        for i in self._registry.get(ResultType.LIBRARY.value, []):
            if dict(**i) not in libs:
                libs.append(i)
        return libs



"""
Run module.
"""
from __future__ import annotations

import shutil
import typing
from pathlib import Path
from typing import Any

from nefertem_core.readers.builder import build_reader
from nefertem_core.run.status import RunStatus
from nefertem_core.stores.builder import get_all_input_stores, get_input_store, get_output_store
from nefertem_core.utils.commons import FILE_READER
from nefertem_core.utils.logger import LOGGER
from nefertem_core.utils.utils import get_time, listify

if typing.TYPE_CHECKING:
    from nefertem_core.run.handler import RunHandler
    from nefertem_core.run.run_info import RunInfo


class Run:
    """
    Run object.
    The Run is the main interface to interact with data, metadata and operational framework.
    With the Run object you can:

    - Execute plugins
    - Log metadata
    - Persist artifacts
    - Persist input data

    Attributes
    ----------
    run_info : RunInfo
        Run information.
    run_handler : RunHandler
        Run handler.
    _tmp_dir : str
        Local temporary folder where to store input data.

    Methods
    -------
    persist_data
        Persist input data as artifacts into default store.

    """

    def __init__(self, run_info: RunInfo, run_handler: RunHandler, tmp_dir: str) -> None:
        """
        Constructor.
        """
        self.run_info = run_info
        self.run_handler = run_handler
        self._tmp_dir = tmp_dir

    ############################
    # Run methods
    ############################

    def _get_base_args(self) -> tuple:
        """
        Return base arguments for metadata.

        Returns
        -------
        tuple
            Base arguments for metadata.
        """
        return self.run_info.run_id, self.run_info.experiment_name

    def _log_run(self) -> None:
        """
        Log run's metadata.

        Returns
        -------
        None
        """
        metadata = self.run_info.to_dict()
        self._log_metadata(metadata, "run_metadata.json")

    def _log_metadata(self, obj: dict, filename: str) -> None:
        """
        Log metadata dictionary.

        Parameters
        ----------
        obj : dict
            Metadata dictionary.
        filename : str
            Metadata filename.

        Returns
        -------
        None
        """
        pth = get_output_store().log_metadata(obj, filename)
        if pth not in self.run_info.output_files:
            self.run_info.output_files.append(pth)

    def _persist_artifact(self, obj: Any, filename: str) -> None:
        """
        Persist artifact in the output store.

        Parameters
        ----------
        obj : Any
            Artifact to persist.
        filename : str
            Artifact filename.

        Returns
        -------
        None
        """
        pth = get_output_store().persist_artifact(obj, filename)
        if pth not in self.run_info.output_files:
            self.run_info.output_files.append(pth)

    ############################
    # Data
    ############################

    def persist_data(self) -> None:
        """
        Persist input data as artifacts.

        Returns
        -------
        None
        """
        for res in self.run_info.resources:
            store = get_input_store(res.store)
            data_reader = build_reader(FILE_READER, store)
            for path in listify(res.path):
                tmp = Path(data_reader.fetch_data(path))
                self._persist_artifact(tmp, tmp.name)

    def _clean_all(self) -> None:
        """
        Clean up.
        """
        for store in get_all_input_stores():
            store.clean_paths()
        try:
            shutil.rmtree(self._tmp_dir)
        except FileNotFoundError:
            pass

    ############################
    # Context manager
    ############################

    def __enter__(self) -> Run:
        # Handle run's start
        LOGGER.info(f"Starting run {self.run_info.run_id}")

        # Set run's status
        self.run_info.status = RunStatus.RUNNING.value
        self.run_info.started = get_time()

        # Log run's metadata
        self._log_run()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        # Handle run's end
        if exc_type is None:
            self.run_info.status = RunStatus.FINISHED.value
        elif exc_type in (InterruptedError, KeyboardInterrupt):
            self.run_info.status = RunStatus.INTERRUPTED.value
        else:
            self.run_info.status = RunStatus.ERROR.value
        self.run_info.finished = get_time()

        # Get libraries used in the run
        self.run_info.run_libraries = self.run_handler.get_libraries()

        # Log run's metadata
        self._log_run()

        # Clean up
        LOGGER.info("Run finished. Clean up of temp resources.")
        self._clean_all()

    ############################
    # Dunder
    ############################

    def __repr__(self) -> str:
        return str(self.run_info.to_dict())



from __future__ import annotations

import typing
from abc import abstractmethod

if typing.TYPE_CHECKING:
    from nefertem_core.plugins.plugin import Plugin
    from nefertem_core.stores.input.objects._base import InputStore


class PluginBuilder:
    """
    Abstract PluginBuilder class.
    """

    def __init__(self, stores: list[InputStore], exec_args: dict) -> None:
        self.exec_args = exec_args
        self.stores = {store.name: store for store in stores}

    @abstractmethod
    def build(self, *args, **kwargs) -> list[Plugin]:
        """
        Build a list of plugin.
        """



"""
Base abstract metadata store module.
"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Any


class OutputStore(metaclass=ABCMeta):
    """
    Abstract class that defines methods on how to persist metadata and artifacts into
    different storage backends.

    Attributes
    ----------
    path : str
        An URI string that points to the storage.

    """

    def __init__(self, path: str) -> None:
        self.path = path

    ############################
    # Run methods
    ############################

    @abstractmethod
    def init_run(self, exp_name: str, run_id: str, overwrite: bool) -> None:
        """
        Initial enviroment operation.
        """

    @abstractmethod
    def get_run_path(self) -> str:
        """
        Return run path.
        """

    ############################
    # Write methods
    ############################

    @abstractmethod
    def log_metadata(self, obj: dict, filename: str) -> None:
        """
        Method to log metadata.
        """

    @abstractmethod
    def persist_artifact(self, obj: Any, filename: str) -> None:
        """
        Method to persist an artifact.
        """



"""
OutputStore registry.
"""
from __future__ import annotations

import typing

from nefertem_core.stores.kinds import StoreKinds
from nefertem_core.stores.output.objects.dummy import DummyOutputStore
from nefertem_core.stores.output.objects.local import LocalOutputStore

if typing.TYPE_CHECKING:
    from nefertem_core.stores.output.objects._base import OutputStore


class OutputStoreRegistry(dict):
    """
    Generic registry for OutputStore objects.
    """

    def register(self, kind: str, store: OutputStore) -> None:
        """
        Register a new store.

        Parameters
        ----------
        kind : str
            The store kind.
        store : OutputStore
            The store object.
        """
        self[kind] = store


mdstore_registry = OutputStoreRegistry()
mdstore_registry.register(StoreKinds.DUMMY.value, DummyOutputStore)
mdstore_registry.register(StoreKinds.LOCAL.value, LocalOutputStore)



"""
Plugin builder factory module.
"""
from __future__ import annotations

import importlib
import typing

from nefertem_core.utils.exceptions import RunError

if typing.TYPE_CHECKING:
    from nefertem_core.plugins.builder import PluginBuilder
    from nefertem_core.run.config import RunConfig


def builder_factory(config: RunConfig, stores: dict) -> list:
    """
    Factory method that creates plugin builders.
    """
    builders = []
    for cfg in config.exec_config:
        ClsBuilder = _get_object(config.operation, cfg.framework)
        builders.append(ClsBuilder(stores, cfg.exec_args))
    return builders


def _get_object(operation: str, framework: str) -> PluginBuilder:
    """
    Get run handler class.

    Parameters
    ----------
    operation : str
        Operation to perform.
    framework : str
        Framework to use.

    Returns
    -------
    PluginBuilder
        Plugin builder class.
    """
    try:
        module = importlib.import_module(f"nefertem_{operation}_{framework}")
        return getattr(module, "Builder")
    except (ImportError, AttributeError):
        raise RunError(f"Builder of {framework} for {operation} not found.")



"""
Data resource module.
"""
from __future__ import annotations

from typing import Union

from nefertem_core.utils.utils import build_uuid
from pydantic import BaseModel, Field


class DataResource(BaseModel):
    """
    Resource configuration class.
    This object represents a physical resource present
    on a backend or a virtual resource rebuildable starting
    from other resources.
    """

    id: str = Field(default_factory=build_uuid)
    """UUID of DataResource."""

    name: str
    """Name of the DataResource."""

    path: Union[str, list[str]]
    """An URI (or a list of URI) that point to data."""

    store: str
    """Store name where to find the resource."""

    package: str = None
    """Package name that DataResource belongs to."""

    title: str = None
    """Human readable name for the DataResource."""

    description: str = None
    """A description of the DataResource."""

    table_schema: Union[str, dict] = None
    """Resource table schema or path to table schema."""



"""
Nefertem base report module.
"""
from __future__ import annotations


class NefertemBaseReport:
    """
    Nefertem base report class.

    Attributes
    ----------
    framework_name : str
        Execution library name.
    framework_version : str
        Execution library version.
    duration : float
        Time required by the execution process.
    """

    def __init__(
        self,
        framework_name: str,
        framework_version: str,
        duration: float,
    ) -> None:
        """
        Constructor.
        """
        super().__init__()
        self.framework_name = framework_name
        self.framework_version = framework_version
        self.duration = duration

    def to_dict(self) -> dict:
        return self.__dict__

    def __repr__(self) -> str:
        return f"{self.to_dict()}"



from __future__ import annotations

import importlib
import typing

from nefertem_core.readers.registry import reader_registry

if typing.TYPE_CHECKING:
    from nefertem_core.readers.objects._base import DataReader
    from nefertem_core.stores.input.objects._base import InputStore


def build_reader(reader_type: str, store: InputStore, **kwargs) -> DataReader:
    """
    Reader builder.

    Parameters
    ----------
    reader_type: str
        Reader type.
    store: InputStore
        Store to read from.
    kwargs: dict
        Reader kwargs.

    Returns
    -------
    DataReader
        Reader instance.
    """
    try:
        module = importlib.import_module(reader_registry[reader_type][0])
        reader = getattr(module, reader_registry[reader_type][1])
        return reader(store, **kwargs)
    except (KeyError, ModuleNotFoundError, AttributeError, ImportError):
        raise KeyError(f"Reader {reader_type} not found. Check installed libraries.")



"""
NativeReader module.
"""
from __future__ import annotations

from typing import Any

from nefertem_core.readers.objects._base import DataReader


class NativeReader(DataReader):
    """
    NativeReader class.

    The NativeReader object tells the stores to return
    a string reference to a resource according to store type,
    e.g. a connection string for SQL store or an encoded URL
    for remote storages.
    """

    def fetch_data(self, src: str) -> Any:
        """
        Fetch resource from backend.
        """
        return self.store.fetch_native(src)



"""
Implementation of SQL artifact store.
"""
from __future__ import annotations

import re
from pathlib import Path

import polars as pl
from nefertem_core.stores.input.objects._base import InputStore, StoreConfig
from nefertem_core.utils.exceptions import StoreError


class SQLStoreConfig(StoreConfig):
    """
    SQL store configuration class.
    """

    driver: str
    """SQL driver."""

    host: str
    """SQL host."""

    port: int
    """SQL port."""

    user: str
    """SQL user."""

    password: str
    """SQL password."""

    database: str
    """SQL database name."""


class SQLInputStore(InputStore):
    """
    SQL artifact store object.

    Allows the client to interact with SQL based storages.

    """

    def __init__(self, name: str, store_type: str, temp_dir: str, config: SQLStoreConfig) -> None:
        """
        Constructor.
        """
        super().__init__(name, store_type, temp_dir)
        self.config = config

    def persist_artifact(self, *args) -> None:
        """
        Persist an artfact.

        Raises
        ------
        NotImplementedError
            This method is not implemented.
        """
        raise NotImplementedError("SQL store does not support persistence.")

    def fetch_file(self, src: str) -> Path:
        """
        Return the path where a resource it is stored.

        Parameters
        ----------
        src : str
            The name of the file.

        Returns
        -------
        str
            The location of the requested file.
        """
        key = f"{src}_file"
        cached = self._get_resource(key)
        if cached is not None:
            return cached

        self.logger.info(f"Fetching resource {src} from store {self.name}")
        table = self._get_table_name(src)
        dst = self.temp_dir / f"{table}.parquet"
        filepath = self._download_table(table, dst)
        self._register_resource(key, filepath)
        return filepath

    def fetch_native(self, *args) -> str:
        """
        Fetch a native resource.

        Returns
        -------
        str
            The connection string.
        """
        return self._get_connection_string()

    ############################
    # Private helper methods
    ############################

    def _get_connection_string(self) -> str:
        """
        Get the connection string.

        Returns
        -------
        str
            The connection string.
        """
        return (
            f"{self.config.driver}//{self.config.user}:{self.config.password}@"
            f"{self.config.host}:{self.config.port}/{self.config.database}"
        )

    @staticmethod
    def _parse_path(path: str) -> dict:
        """
        Parse the path and return the components.

        Parameters
        ----------
        path : str
            The path.

        Returns
        -------
        dict
            A dictionary containing the components of the path.
        """
        pattern = r"^sql://(?P<database>.+)/(?P<table>.+)$"
        match = re.match(pattern, path)
        if match is None:
            raise ValueError("Invalid SQL path. Must be sql://<database>/<table>")
        return match.groupdict()

    def _get_table_name(self, uri: str) -> str:
        """
        Get the name of the table from the URI.

        Parameters
        ----------
        uri : str
            The URI.

        Returns
        -------
        str
            The name of the table
        """
        return str(self._parse_path(uri).get("table"))

    ############################
    # Private I/O methods
    ############################

    def _execute_query(self, query: str) -> pl.DataFrame:
        """
        Execute a query.

        Parameters
        ----------
        query : str
            The query.

        Returns
        -------
        list
            The query results.
        """
        pl.read_database(query, self._get_connection_string(), engine="adbc")

    def _download_table(self, table: str, dst: str) -> Path:
        """
        Download a table from SQL based storage.

        Parameters
        ----------
        table : str
            The origin table.
        dst : str
            The destination path.

        Returns
        -------
        Path
            The destination of the file on local filesystem.
        """
        self._verify_table(table)
        query = f"select * from {table}"
        self._execute_query(query).write_parquet(dst)
        return Path(dst)

    def _verify_table(self, table: str) -> None:
        """
        Verify if table exists.

        Parameters
        ----------
        table : str
            Table name.

        Returns
        -------
        None
        """
        query = """
        SELECT TABLE_NAME as table
        FROM INFORMATION_SCHEMA.TABLES
        """
        res = self._execute_query(query)
        if table not in res.to_series().to_list():
            raise StoreError(f"Table {table} not in db.")



"""
Common IO utils.
"""
from __future__ import annotations

import json
from io import BytesIO, StringIO
from pathlib import Path


def write_bytesio(src: str) -> BytesIO:
    """
    Write string in ByteStream BytesIO.

    Parameters
    ----------
    src : str
        The source string to be wrapped.

    Returns
    -------
    BytesIO
        The wrapped object.
    """
    bytesio = BytesIO()
    bytesio.write(src.encode())
    bytesio.seek(0)
    return bytesio


def write_bytes(byt: bytes, path: Path) -> None:
    """
    Write bytes on a file.

    Parameters
    ----------
    byt : bytes
        The bytes to be written.
    path : Path
        The path to the file.

    Returns
    -------
    None
    """
    path.write_bytes(byt)


def write_object(buff: BytesIO | StringIO, path: Path) -> None:
    """
    Write a buffer as file.

    Parameters
    ----------
    buff : IO
        The buffer to be written.
    path : Path
        The path to the file.

    Returns
    -------
    None
    """
    buff.seek(0)
    if isinstance(buff, BytesIO):
        path.write_bytes(buff.read())
    else:
        path.write_text(buff.read(), encoding="utf-8")


def write_json(data: dict, path: Path) -> None:
    """
    Store JSON file.

    Parameters
    ----------
    data : dict
        The data to be stored.
    path : Path
        The path to the file.

    Returns
    -------
    None
    """
    path.write_text(json.dumps(data, indent=4), encoding="utf-8")



"""
Local metadata store module.
"""
from __future__ import annotations

import shutil
from io import BytesIO, StringIO
from pathlib import Path
from typing import Any

from nefertem_core.stores.output.objects._base import OutputStore
from nefertem_core.utils.exceptions import RunError
from nefertem_core.utils.io_utils import write_json, write_object


class LocalOutputStore(OutputStore):
    """
    Local metadata store object.

    Allows the client to interact with local filesystem.
    """

    def __init__(self, path: str) -> None:
        super().__init__(path)

        self.path = Path(self.path)

        self._initialized = False
        self._run_path = None
        self._artifact_path = None
        self._metadata_path = None

    ############################
    # Run methods
    ############################

    def init_run(self, exp_name: str, run_id: str, overwrite: bool) -> None:
        """
        Initial run folders. If folder doesn't exist, create it.
        If overwrite is True, delete the folder and create it again.

        Parameters
        ----------
        exp_name : str
            Experiment name.
        run_id : str
            Run id.

        Returns
        -------
        None
        """
        self._set_paths(exp_name, run_id)
        if self._run_path.exists():
            if not overwrite:
                raise RunError("Run already exists, please use another id.")
            else:
                shutil.rmtree(self._run_path)
                self._create_run_directories()
        else:
            self._create_run_directories()
        self._initialized = True

    def _set_paths(self, exp_name: str, run_id: str) -> None:
        """
        Set run paths.

        Parameters
        ----------
        exp_name : str
            Experiment name.
        run_id : str
            Run id.

        Returns
        -------
        None
        """
        self._run_path = self.path / exp_name / run_id
        self._artifact_path = self._run_path / "artifacts"
        self._metadata_path = self._run_path / "metadata"

    def _create_run_directories(self) -> None:
        """
        Create run folders.

        Returns
        -------
        None
        """
        self._run_path.mkdir(parents=True)
        self._artifact_path.mkdir()
        self._metadata_path.mkdir()

    def get_run_path(self) -> Path:
        """
        Return run path.

        Parameters
        ----------
        exp_name : str
            Experiment name.
        run_id : str
            Run id.

        Returns
        -------
        Path
            Run path.

        Raises
        ------
        RunError
            If the run is not initialized.
        """
        if not self._initialized:
            raise RunError("Run not initialized.")
        return self._run_path

    ############################
    # Write methods
    ############################

    def log_metadata(self, obj: dict, filename: str) -> Path:
        """
        Method that log metadata.

        Parameters
        ----------
        obj: dict
            Metadata dictionary to be logged.
        filename: str
            Filename for the metadata.

        Returns
        -------
        Path
            Path to the metadata file.
        """

        if not isinstance(obj, dict):
            raise RunError("Metadata must be a dictionary.")
        dst = self._metadata_path / filename
        write_json(obj, dst)
        return dst

    def persist_artifact(self, obj: Any, filename: str) -> Path:
        """
        Method to persist an artifact.
        The local store supports the following types:

        - Local file or dump string
        - Dictionary
        - StringIO/BytesIO buffer

        Parameters
        ----------
        obj : Any
            The source object to be persisted.
        filename: str
            Filename for the artifact.

        Returns
        -------
        str
            Path to the artifact.

        Raises
        ------
        RunError
            If the source type is not supported.
        """
        dst = self._artifact_path / filename

        if isinstance(obj, (str, Path)):
            shutil.copy(obj, dst)

        elif isinstance(obj, dict):
            write_json(obj, dst)

        elif isinstance(obj, (BytesIO, StringIO)):
            write_object(obj, dst)

        else:
            raise RunError("Invalid object type, it can not be persisted.")

        return dst



"""
Run configuration objects module.
"""
from __future__ import annotations

from pydantic import BaseModel


class ExecConfig(BaseModel):
    """
    Generic configuration for run operation.
    """

    framework: str
    """Library to use for performing an operation."""

    exec_args: dict = {}
    """Execution arguments to pass to plugins."""


class RunConfig(BaseModel):
    """
    Run configuration.
    """

    operation: str
    """Operation to perform."""

    exec_config: list[ExecConfig]
    """Execution configuration."""

    parallel: bool = False
    """Flag to execute operation in parallel."""

    num_worker: int = 10
    """Number of workers to execute operation in parallel, by default 10"""



"""
Module for common terms definition.
"""

# Nefertem version
NEFERTEM_VERSION: str = "2.0.0"

# Data readers type
FILE_READER: str = "file_reader"
NATIVE_READER: str = "native_readerr"

# Generics
DUMMY: str = "_dummy"



"""
Common generic utils.
"""
from __future__ import annotations

import functools
import operator
from datetime import datetime
from typing import Any
from uuid import uuid4


def build_uuid(uuid: str | None = None) -> str:
    """
    Create a uuid if not given

    Parameters
    ----------
    uuid : str
        UUID. Optional.

    Returns
    -------
    str
        The uuid.
    """
    if uuid is not None:
        return uuid
    return str(uuid4())


def flatten_list(list_of_list: list[list[Any]]) -> list[Any]:
    """
    Flatten a list of list.

    Parameters
    ----------
    list_of_list : list[list[Any]]
        The list of list to be flattened.
    """
    try:
        return functools.reduce(operator.iconcat, list_of_list)
    except TypeError:
        return []


def listify(obj: Any) -> list[Any]:
    """
    Convert an object to a list.

    Parameters
    ----------
    obj : Any
        The object to be converted.

    Returns
    -------
    list[Any]
        The list.
    """
    if not isinstance(obj, (list, tuple)):
        obj = [obj]
    return obj


def get_time() -> str:
    """
    Get current time with timezone info.

    Returns
    -------
    str
        ISO 8601 time with timezone info.
    """
    return datetime.now().astimezone().isoformat(timespec="milliseconds")



"""
Dummy artifact store module.
"""
from __future__ import annotations

from nefertem_core.stores.input.objects._base import InputStore, StoreConfig


class DummyStoreConfig(StoreConfig):
    """
    Dummy store configuration.
    """


class DummyInputStore(InputStore):
    """
    Dummy artifact store object implementation.

    Only allows the client to interact store methods.
    """

    def __init__(
        self,
        name: str,
        store_type: str,
        temp_dir: str,
        config: DummyStoreConfig,
    ) -> None:
        """
        Constructor.
        """
        super().__init__(name, store_type, temp_dir)
        self.config = config

    ############################
    # Read methods
    ############################

    def fetch_file(self, *args) -> None:
        """
        Placeholder methods.

        Returns
        -------
        None
        """

    def fetch_native(self, *args) -> None:
        """
        Placeholder methods.

        Returns
        -------
        None
        """



"""
StoreBuilder module.
"""
from __future__ import annotations

import typing

from nefertem_core.stores.input.objects._base import StoreParameters
from nefertem_core.stores.input.registry import input_store_registry
from nefertem_core.stores.kinds import StoreKinds
from nefertem_core.stores.output.registry import mdstore_registry
from nefertem_core.utils.commons import DUMMY
from nefertem_core.utils.exceptions import StoreError
from pydantic import ValidationError

if typing.TYPE_CHECKING:
    from nefertem_core.stores.input.objects._base import InputStore, StoreConfig
    from nefertem_core.stores.output.objects._base import OutputStore


class StoreBuilder:
    """
    StoreBuilder class.
    """

    def __init__(self) -> None:
        """
        Constructor.
        """
        self._stores: dict = {}
        self._output_store: OutputStore | None = None

    def build_output_store(self, path: str | None = None) -> None:
        """
        Method to create an output stores. If the path is None, the method creates a dummy
        output store.

        Parameters
        ----------
        path: str
            Output path.

        Returns
        -------
        None
        """
        if path is None:
            return mdstore_registry[StoreKinds.DUMMY.value](DUMMY)
        if self._output_store is None:
            self._output_store = mdstore_registry[StoreKinds.LOCAL.value](path)

    def build_input_store(self, temp_dir: str, config: dict | None = None) -> None:
        """
        Method to create an input stores.

        Parameters
        ----------
        config : dict
            Store configuration.
        temp_dir: str
            Temporary directory.

        Returns
        -------
        None
        """
        params = self._parse_parameters(temp_dir, config)
        if params["name"] in self._stores:
            raise StoreError(f"Store {params['name']} already exists.")
        self._stores[params["name"]] = self._get_store(params)

    def _parse_parameters(self, temp_dir: str, config: dict | None = None) -> dict:
        """
        Parse store parameters.

        Parameters
        ----------
        temp_dir : str
            Temporary directory.
        config : dict
            Store configuration.

        Returns
        -------
        dict
            Store parameters.
        """
        cfg: StoreParameters = self._validate_parameters(config)
        return {
            "name": cfg.name,
            "store_type": cfg.store_type,
            "temp_dir": temp_dir,
            "config": self._validate_config(cfg.store_type, cfg.config),
        }

    @staticmethod
    def _validate_parameters(config: dict | None = None) -> StoreParameters:
        """
        Validate store parameters against a pydantic model.

        Parameters
        ----------
        config : dict
            Store configuration.

        Returns
        -------
        StoreParameters
            Store parameters.

        Raises
        ------
        StoreError
            If the store configuration is invalid.
        """
        try:
            return StoreParameters(**config)
        except TypeError:  # If config is None
            return StoreParameters(StoreKinds.DUMMY.value, DUMMY)
        except ValidationError:
            raise StoreError("Invalid store configuration.")

    @staticmethod
    def _validate_config(store_type: str, config: dict | None = None) -> StoreConfig:
        """
        Validate store configuration.

        Parameters
        ----------
        store_type : str
            Store type.
        config : dict
            Store configuration.

        Returns
        -------
        dict
            Store configuration.

        Raises
        ------
        StoreError
            If the store configuration is invalid.
        """
        try:
            return input_store_registry[store_type]["model"](**config)
        except (ValidationError, TypeError):
            raise StoreError("Invalid store configuration.")
        except KeyError:
            raise StoreError("Invalid store type.")

    @staticmethod
    def _get_store(params: dict) -> InputStore:
        """
        Validate store configuration.

        Parameters
        ----------
        params : dict
            Store configuration.

        Returns
        -------
        dict
            Store configuration.

        Raises
        ------
        StoreError
            If the store configuration is invalid.
        """
        try:
            store_type = params["store_type"]
            return input_store_registry[store_type]["store"](**params)
        except TypeError:
            raise StoreError("Something went wrong.")
        except KeyError:
            raise StoreError("Invalid store type.")

    def get_input_store(self, name: str) -> InputStore:
        """
        Get store by name.

        Parameters
        ----------
        name : str
            Store name.

        Returns
        -------
        InputStore
            Artifact store object.

        Raises
        ------
        StoreError
            If the store is not found.
        """
        store = self._stores.get(name)
        if store is None:
            raise StoreError(f"Store {name} not found.")
        return store

    def get_all_input_stores(self) -> list[InputStore]:
        """
        Get all stores.

        Returns
        -------
        list
            List of artifact store objects.
        """
        stores = list(self._stores.values())
        if not stores:
            raise StoreError("No stores found.")
        return stores

    def get_output_store(self) -> OutputStore:
        """
        Get output store.

        Returns
        -------
        OutputStore
            Metadata store object.

        Raises
        ------
        StoreError
            If the store is not found.
        """
        if self._output_store is None:
            raise StoreError("Output store not found.")
        return self._output_store


store_builder = StoreBuilder()


def get_input_store(name: str) -> InputStore:
    """
    Wrapper for StoreBuilder.get_input_store.

    Parameters
    ----------
    name : str
        Store name.

    Returns
    -------
    InputStore
        Artifact store object.
    """
    return store_builder.get_input_store(name)


def get_all_input_stores() -> list[InputStore]:
    """
    Wrapper for StoreBuilder.get_all_input_stores.

    Returns
    -------
    list
        List of artifact store objects.
    """
    return store_builder.get_all_input_stores()


def get_output_store() -> OutputStore:
    """
    Wrapper for StoreBuilder.get_output_store.

    Returns
    -------
    OutputStore
        Metadata store object.
    """
    return store_builder.get_output_store()



"""
DataReader registry.
"""
from __future__ import annotations

from nefertem_core.utils.commons import FILE_READER, NATIVE_READER


class ReaderRegistry(dict):
    def register(self, name: str, module: str, class_name: str) -> None:
        """
        Register a reader.

        Parameters
        ----------
        name : str
            Reader name.
        module : str
            Reader module.
        class_name : str
            Reader class name.

        Returns
        -------
        None
        """
        self[name] = [module, class_name]


reader_registry = ReaderRegistry()
reader_registry.register(FILE_READER, "nefertem_core.readers.objects.file", "FileReader")
reader_registry.register(NATIVE_READER, "nefertem_core.readers.objects.native", "NativeReader")



"""
Stores kinds module.
"""
from enum import Enum


class StoreKinds(Enum):
    """
    Enumerates the available store kinds.
    """

    DUMMY = "_dummy"
    LOCAL = "local"
    REMOTE = "remote"
    S3 = "s3"
    SQL = "sql"



"""
Base abstract Run Plugin module.
"""
from __future__ import annotations

import typing
from abc import ABCMeta, abstractmethod

from nefertem_core.plugins.utils import RenderTuple
from nefertem_core.utils.logger import LOGGER
from nefertem_core.utils.utils import build_uuid

if typing.TYPE_CHECKING:
    from nefertem_core.plugins.utils import Result


class Plugin(metaclass=ABCMeta):
    """
    Base plugin abstract class.
    """

    def __init__(self) -> None:
        self.id = build_uuid()
        self.logger = LOGGER

        self.data_reader = None
        self.exec_args = None

        self.exec_sequential = True
        self.exec_multiprocess = False
        self.exec_multithread = False

    @abstractmethod
    def setup(self, *args, **kwargs) -> None:
        """
        Configure a plugin.
        """

    @abstractmethod
    def execute(self) -> dict:
        """
        Execute main plugin operation.
        """

    @abstractmethod
    def render_nefertem(self, obj: Result) -> RenderTuple:
        """
        Return a NefertemProfile object ready to be persisted as metadata.
        """

    @abstractmethod
    def render_artifact(self, obj: Result) -> list[RenderTuple]:
        """
        Return an object ready to be persisted as artifact.
        """

    @staticmethod
    @abstractmethod
    def framework_name() -> str:
        """
        Get framework name.
        """

    @staticmethod
    @abstractmethod
    def framework_version() -> str:
        """
        Get framework version.
        """

    def get_framework(self) -> dict:
        """
        Get framework info.
        """
        return {
            "framework_name": self.framework_name(),
            "framework_version": self.framework_version(),
        }



"""
Abstract class for artifact store.
"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Any

from nefertem_core.utils.logger import LOGGER
from nefertem_core.utils.utils import build_uuid
from pydantic import BaseModel


class StoreConfig(BaseModel):
    """
    Store configuration class.
    It defines the configuration settings for a Store like endpoints, credentials, etc.
    It is modelled on different kind of Store.
    """


class StoreParameters(BaseModel):
    """
    Store configuration class.
    """

    name: str
    """Store id."""

    store_type: str
    """Store type."""

    config: dict = {}
    """Dictionary containing the configuration for the backend."""


class InputStore(metaclass=ABCMeta):
    """
    Abstract class that defines methods to fetch artifacts from different storage backends.

    Attributes
    ----------
    name : str
        Name of store.
    store_type : str
        Type of store (local, remote, s3, sql).
    temp_dir : str
        Temporary download path.
    config : dict, default = None
        A dictionary with the credentials/configurations for the storage.
    """

    def __init__(self, name: str, store_type: str, temp_dir: str) -> None:
        """
        Constructor.
        """
        self.name = name
        self.store_type = store_type
        self.temp_dir = Path(temp_dir) / build_uuid()

        # Path registry
        self._cache = {}

        # Logger
        self.logger = LOGGER

    ############################
    # Read methods
    ############################

    @abstractmethod
    def fetch_file(self, src: str) -> Path:
        """
        Return the temporary path where a resource it is stored.
        """

    @abstractmethod
    def fetch_native(self, src: str) -> Any:
        """
        Return a native format path for a resource.
        """

    ############################
    # Cache methods
    ############################

    def _get_resource(self, key: str) -> Path | None:
        """
        Method to return temporary path of a registered resource.
        """
        return self._cache.get(key)

    def _register_resource(self, key: str, path: Path) -> None:
        """
        Method to register a resource into the path registry.
        """
        if key not in self._cache:
            self._cache[key] = path

    def clean_paths(self) -> None:
        """
        Delete all temporary paths references from stores.
        """
        self._cache = {}



"""
Blob module.
"""
from __future__ import annotations

from nefertem_core.utils.commons import NEFERTEM_VERSION


class Blob:
    """
    Object logged to backend.

    Attributes
    ----------
    run_id : str
        Run id.
    experiment_name : str
        Experiment name.
    contents : dict
        Blob of metadata to log.
    """

    def __init__(self, run_id: str, experiment_name: str, contents: dict) -> None:
        """
        Constructor.
        """
        self.run_id = run_id
        self.experiment_name = experiment_name
        self.contents = contents
        self.nefertem_version = NEFERTEM_VERSION

    def to_dict(self) -> dict:
        """
        Render the object as a dictionary.

        Returns
        -------
        dict
            Dictionary representation of the object.
        """
        return self.__dict__



"""
Implementation of REST artifact store.
"""
from __future__ import annotations

from pathlib import Path
from urllib.parse import unquote, urlparse

import requests
from nefertem_core.stores.input.objects._base import InputStore, StoreConfig


class RemoteStoreConfig(StoreConfig):
    """
    HTTP store configuration class.
    """

    auth: str = None
    """Authentication type."""

    user: str = None
    """User name."""

    password: str = None
    """Password."""

    token: str = None
    """Bearer token."""


class RemoteInputStore(InputStore):
    """
    Rest artifact store object.

    Allows the client to interact with remote HTTP store.

    """

    def __init__(self, name: str, store_type: str, temp_dir: str, config: RemoteStoreConfig) -> None:
        """
        Constructor.
        """
        super().__init__(name, store_type, temp_dir)
        self.config = config

    ############################
    # Read methods
    ############################

    def fetch_file(self, src: str) -> Path:
        """
        Return the path where a resource it is stored.

        Parameters
        ----------
        src : str
            The name of the file.

        Returns
        -------
        Path
            The location of the requested file.
        """
        key = f"{src}_file"
        cached = self._get_resource(key)
        if cached is not None:
            return cached

        self.logger.info(f"Fetching resource {src} from store {self.name}")
        dst = self.temp_dir / self._get_filename(src)
        filepath = self._download_file(src, dst)
        self._register_resource(key, filepath)
        return filepath

    def fetch_native(self, src: str) -> str:
        """
        Return a native format path for a resource.

        Parameters
        ----------
        src : str
            The URL of the resource.

        Returns
        -------
        str
            The URL of the resource.
        """
        return src

    ############################
    # Helper methods
    ############################

    @staticmethod
    def _check_head(src) -> None:
        """
        Check if the source exists.

        Parameters
        ----------
        src : str
            The source location.

        Returns
        -------
        None

        Raises
        ------
        HTTPError
            If an error occurs while checking the source.
        """
        r = requests.head(src, timeout=60)
        r.raise_for_status()

    def _download_file(self, url: str, dst: str) -> Path:
        """
        Method to download a file from a given url.

        Parameters
        ----------
        url : str
            The url of the file to download.
        dst : str
            The destination of the file.

        Returns
        -------
        Path
            The path of the downloaded file.
        """
        self._check_head(url)
        kwargs = self._get_auth()
        with requests.get(url, stream=True, timeout=60, **kwargs) as r:
            r.raise_for_status()
            with open(dst, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
        return Path(dst)

    def _get_auth(self) -> dict:
        """
        Get authentication parameters from the config.

        Returns
        -------
        dict
            The authentication parameters.
        """
        if self.config.auth == "basic":
            return {"auth": (self.config.user, self.config.password)}
        if self.config.auth == "oauth":
            return {"headers": {"Authorization": f"Bearer {self.config.token}"}}
        return {}

    @staticmethod
    def _get_filename(url: str) -> str:
        """
        Get the filename from a url.

        Parameters
        ----------
        url : str
            The url of the file.

        Returns
        -------
        str
            The filename.

        Raises
        ------
        ValueError
            If the file extension is not supported.
        """
        filename = Path(unquote(urlparse(url).path)).name
        if filename.split(".")[-1] not in ["csv", "parquet"]:
            raise ValueError("Only csv and parquet files are supported for download.")
        return filename



"""
RunInfo module.
"""
from __future__ import annotations

import typing
from pathlib import Path

from nefertem_core.metadata.env import Env
from nefertem_core.run.status import RunStatus
from nefertem_core.utils.commons import NEFERTEM_VERSION

if typing.TYPE_CHECKING:
    from nefertem_core.resources.data_resource import DataResource
    from nefertem_core.run.config import RunConfig


class RunInfo:
    """
    Run's metadata.

    Attributes
    ----------
    run_id : str
        Run id.
    experiment_name : str
        Id of the experiment.
    run_path : Path
        Path where to store run's metadata and artifacts.
    resources : list[DataResource]
        List of input resources.
    run_config : RunConfig
        Run configuration.
    """

    def __init__(
        self,
        run_id: str,
        experiment_name: str,
        run_path: Path,
        run_config: RunConfig,
        resources: list[DataResource],
    ) -> None:
        """
        Constructor.
        """

        # Run info
        self.run_id = run_id
        self.experiment_name = experiment_name

        # Execution info
        self.run_path = run_path
        self.run_config = run_config
        self.run_libraries = None

        # Inputs
        self.resources = resources

        # Outputs
        self.output_files = []

        # Execution environment
        self.nefertem_version = NEFERTEM_VERSION
        self.execution_environment = Env()

        # Status
        self.status = RunStatus.CREATED.value

        # Timings
        self.started = None
        self.finished = None

    def to_dict(self) -> dict:
        """
        Override the method to_dict of the Metadata class.

        Returns
        -------
        dict
            Dictionary representation of the object.
        """
        return {
            "run_id": self.run_id,
            "experiment_name": self.experiment_name,
            "run_path": str(self.run_path),
            "run_config": self.run_config.dict(exclude_none=True),
            "run_libraries": self.run_libraries,
            "resources": [i.dict(exclude_none=True) for i in self.resources],
            "output_files": [str(i) for i in self.output_files],
            "nefertem_version": self.nefertem_version,
            "execution_environment": self.execution_environment.to_dict(),
            "status": self.status,
            "started": self.started,
            "finished": self.finished,
        }



"""
Run status module.
"""
from __future__ import annotations

from enum import Enum


class RunStatus(Enum):
    """
    Run status.

    Attributes
    ----------
    RUNNING : int
        Run is running.
    COMPLETED : int
        Run is completed.
    FAILED : int
        Run failed.
    """

    CREATED = "created"
    RUNNING = "running"
    FINISHED = "finished"
    INTERRUPTED = "interrupted"
    ERROR = "error"



"""
InputStore registry.
"""
from __future__ import annotations

import typing

from nefertem_core.stores.input.objects.dummy import DummyInputStore, DummyStoreConfig
from nefertem_core.stores.input.objects.local import LocalInputStore, LocalStoreConfig
from nefertem_core.stores.input.objects.remote import RemoteInputStore, RemoteStoreConfig
from nefertem_core.stores.input.objects.s3 import S3InputStore, S3StoreConfig
from nefertem_core.stores.input.objects.sql import SQLInputStore, SQLStoreConfig
from nefertem_core.stores.kinds import StoreKinds

if typing.TYPE_CHECKING:
    from nefertem_core.stores.input.objects._base import InputStore, StoreConfig


class InputStoreRegistry(dict):
    """
    Generic registry for InputStore objects.
    """

    def register(self, kind: str, store: InputStore, model: StoreConfig) -> None:
        """
        Register a new store.

        Parameters
        ----------
        kind : str
            The store kind.
        store : InputStore
            The store object.
        """
        self[kind] = {}
        self[kind]["store"] = store
        self[kind]["model"] = model


input_store_registry = InputStoreRegistry()
input_store_registry.register(StoreKinds.DUMMY.value, DummyInputStore, DummyStoreConfig)
input_store_registry.register(StoreKinds.LOCAL.value, LocalInputStore, LocalStoreConfig)
input_store_registry.register(StoreKinds.S3.value, S3InputStore, S3StoreConfig)
input_store_registry.register(StoreKinds.REMOTE.value, RemoteInputStore, RemoteStoreConfig)
input_store_registry.register(StoreKinds.SQL.value, SQLInputStore, SQLStoreConfig)



"""
Exceptions module.
"""


class NefertemError(Exception):
    """
    Base class for nefertem exception.
    """


class StoreError(NefertemError):
    """
    Raised when incontered errors on Stores.
    """


class RunError(NefertemError):
    """
    Raised when incontered errors on Runs.
    """



"""
FileReader module.
"""
from __future__ import annotations

from pathlib import Path

from nefertem_core.readers.objects._base import DataReader


class FileReader(DataReader):
    """
    FileReader class.

    The FileReader invokes stores fetch_file method to fetch the resource from the backend
    and returns the path to the downloaded resource.
    """

    def fetch_data(self, src: str) -> Path:
        """
        Fetch resource from backend.

        Parameters
        ----------
        src : str
            Resource path.

        Returns
        -------
        Path
            Path to the downloaded resource.
        """
        return self.store.fetch_file(src)



"""
RunBuilder module.
"""
from __future__ import annotations

import importlib
import typing

from nefertem_core.resources.data_resource import DataResource
from nefertem_core.run.config import RunConfig
from nefertem_core.run.handler import RunHandler
from nefertem_core.run.run_info import RunInfo
from nefertem_core.stores.builder import get_output_store
from nefertem_core.utils.exceptions import RunError
from nefertem_core.utils.utils import build_uuid
from pydantic import ValidationError

if typing.TYPE_CHECKING:
    from nefertem_core.run.run import Run


class RunBuilder:
    """
    RunBuilder object to initialize and create runs.

    Attributes
    ----------
    None

    Methods
    -------
    create_run
        Create a new run.
    """

    def create_run(
        self,
        resources: list[dict],
        run_config: dict,
        tmp_dir: str,
        experiment_name: str | None = None,
        run_id: str | None = None,
        overwrite: bool = False,
    ) -> Run:
        """
        Create a new run.

        Parameters
        ----------
        resources : list[dict]
            List of resources to add to the run.
        run_config : dict
            Run configuration.
        tmp_dir : str
            Temporary directory to store artifacts.
        experiment_name : str
            Experiment name.
        run_id : str
            Run id, by default None.
        overwrite : bool
            If True, overwrite run if already exists.

        Returns
        -------
        Run
            Run object.
        """
        if experiment_name is None:
            experiment_name = "default"

        # Validate resources
        res = self._validate_resources(resources)
        self._check_resources(res)

        # Validate run config
        cfg = self._validate_run_config(run_config)

        # Get run id
        run_id = build_uuid(run_id)

        # Initialize run and get run path
        store = get_output_store()
        store.init_run(experiment_name, run_id, overwrite)
        run_path = store.get_run_path()

        # Get run specific operations
        ClsRun: Run = self._get_run_object(cfg.operation)

        # Create run
        run_handler = RunHandler(cfg)
        run_info = RunInfo(
            run_id=run_id,
            experiment_name=experiment_name,
            run_path=run_path,
            run_config=cfg,
            resources=res,
        )
        return ClsRun(run_info, run_handler, tmp_dir)

    @staticmethod
    def _validate_resources(resources: list[dict]) -> list[DataResource]:
        """
        Validate data resources against model.

        Parameters
        ----------
        resources : list[dict]
            List of resources to validate.

        Returns
        -------
        list[DataResource]
            List of validated resources.

        Raises
        ------
        RunError
            If a resource does not match the model.
        """
        try:
            return [DataResource(**res) for res in resources]
        except (TypeError, ValidationError):
            raise RunError("Invalid resource!")

    @staticmethod
    def _check_resources(resources: list[DataResource]) -> None:
        """
        Check if resources already exist.

        Parameters
        ----------
        resources : list[DataResource]
            List of resources to check.

        Returns
        -------
        None

        Raises
        ------
        RunError
            If a resource already exists.
        """
        exists = []
        for res in resources:
            if res.name in exists:
                raise RunError(f"Resource with name {res.name} already exists!")
            exists.append(res.name)

    def _validate_run_config(self, run_config: dict) -> RunConfig:
        """
        Validate run config against model.

        Parameters
        ----------
        run_config : dict
            Run config to validate.

        Returns
        -------
        RunConfig
            Validated run config.

        Raises
        ------
        RunError
            If run config does not match the model.
        """
        try:
            return RunConfig(**run_config)
        except (TypeError, ValidationError):
            raise RunError("Invalid run config!")

    def _get_run_object(self, operation: str) -> Run:
        """
        Get run class.

        Parameters
        ----------
        operation : str
            Operation to perform.

        Returns
        -------
        Run
            Run specific class.

        Raises
        ------
        RunError
            If run class does not exist.
        """
        try:
            module = importlib.import_module(f"nefertem_{operation}.run.run")
            run: Run = getattr(module, f"Run{operation.capitalize()}")
            return run
        except AttributeError:
            raise RunError(f"Run for operation {operation} does not exist!")


run_builder = RunBuilder()



"""
Plugin utils module.
"""
from __future__ import annotations

import time
from collections import namedtuple
from enum import Enum
from typing import Any, Callable

RenderTuple = namedtuple("RenderTuple", ("object", "filename"))


class ExecutionStatus(Enum):
    """
    Enum class for execution status.
    """

    INIT = "created"
    RUNNING = "executing"
    FINISHED = "finished"
    ERROR = "error"


class ResultType(Enum):
    """
    Enum class for result type.
    """

    FRAMEWORK = "framework"
    NEFERTEM = "nefertem"
    RENDERED = "rendered"
    LIBRARY = "library"


class Result:
    """
    Simple class to aggregate result of plugin operation.
    """

    def __init__(
        self,
        status: str | None = ExecutionStatus.INIT.value,
        duration: float | None = None,
        errors: tuple | None = None,
        artifact: Any | None = None,
    ) -> None:
        """
        Constructor.
        """
        self.status = status
        self.duration = duration
        self.errors = errors
        self.artifact = artifact

    def to_dict(self) -> dict:
        """
        Return result as dict.

        Returns
        -------
        dict
            Result as dict.
        """
        return {
            "status": self.status,
            "duration": self.duration,
            "errors": self.errors,
            "artifact": self.artifact,
        }


def exec_decorator(fnc: Callable) -> Result:
    """
    Decorator that keeps track of execution time and status.
    """

    def wrapper(*args, **kwargs) -> Result:
        """
        Wrapper.
        """
        data = Result()
        start = time.perf_counter()
        data.status = ExecutionStatus.RUNNING.value
        try:
            data.artifact = fnc(*args, **kwargs)
            data.status = ExecutionStatus.FINISHED.value
        except Exception as exc:
            data.errors = exc.args
            data.status = ExecutionStatus.ERROR.value
        data.duration = round(time.perf_counter() - start, 2)
        return data

    return wrapper



"""
LocalInputStore module.
"""
from __future__ import annotations

from pathlib import Path

from nefertem_core.stores.input.objects._base import InputStore, StoreConfig


class LocalStoreConfig(StoreConfig):
    """
    Local store configuration class.
    """


class LocalInputStore(InputStore):
    """
    Implementation of a local artifact store object that allows the user to
    interact with the local filesystem.
    """

    def __init__(self, name: str, store_type: str, temp_dir: str, config: LocalStoreConfig) -> None:
        """
        Constructor.
        """
        super().__init__(name, store_type, temp_dir)
        self.config = config

    ############################
    # Read methods
    ############################

    def fetch_file(self, src: str) -> Path:
        """
        Return the path where a resource it is stored.

        Parameters
        ----------
        src : str
            The name of the file.

        Returns
        -------
        Path
            The location of the requested file.
        """
        return Path(src)

    def fetch_native(self, src: str) -> Path:
        """
        Return a native format path for a resource.

        Parameters
        ----------
        src : str
            The name of the file.

        Returns
        -------
        Path
            The location of the requested file.
        """
        return Path(src)



"""
DataReader module.
"""
from __future__ import annotations

import typing
from abc import ABCMeta, abstractmethod
from typing import Any

if typing.TYPE_CHECKING:
    from nefertem_core.stores.input.objects._base import InputStore


class DataReader(metaclass=ABCMeta):
    """
    DataReader abstract class.

    This is the basic abstract class for the DataReaders.

    """

    def __init__(self, store: InputStore) -> None:
        self.store = store

    @abstractmethod
    def fetch_data(self, src: str) -> Any:
        """
        Fetch resources from backend.
        """



"""
Implementation of S3 artifact store.
"""
from __future__ import annotations

# pylint: disable=unused-import
from pathlib import Path
from typing import Type

import boto3
import botocore.client
from botocore.exceptions import ClientError
from nefertem_core.stores.input.objects._base import InputStore, StoreConfig
from nefertem_core.utils.exceptions import StoreError

# Type aliases
S3Client = Type["botocore.client.S3"]


class S3StoreConfig(StoreConfig):
    """
    S3 store configuration class.
    """

    endpoint_url: str
    """S3 endpoint URL."""

    aws_access_key_id: str
    """AWS access key ID."""

    aws_secret_access_key: str
    """AWS secret access key."""

    bucket_name: str
    """S3 bucket name."""


class S3InputStore(InputStore):
    """
    S3 artifact store object.

    Allows the client to interact with S3 based storages.

    """

    def __init__(self, name: str, store_type: str, temp_dir: str, config: S3StoreConfig) -> None:
        """
        Constructor.
        """
        super().__init__(name, store_type, temp_dir)
        self.config = config

    ############################
    # Read methods
    ############################

    def fetch_file(self, src: str) -> str:
        """
        Return the path where a resource it is stored.

        Parameters
        ----------
        src : str
            The name of the file.

        Returns
        -------
        str
            The location of the requested file.
        """
        key = f"{src}_file"
        cached = self._get_resource(key)
        if cached is not None:
            return cached

        self.logger.info(f"Fetching resource {src} from store {self.name}")
        dst = self.temp_dir / src.split("/")[-1]
        filepath = self._download_file(src, dst)
        self._register_resource(key, filepath)
        return filepath

    def fetch_native(self, src: str) -> str:
        """
        Return a native format path for a resource.

        Parameters
        ----------
        src : str
            Key of the resource.

        Returns
        -------
        str
            A presigned URL.
        """
        return self._get_presinged_url(src)

    ############################
    # Private helper methods
    ############################

    def _get_bucket(self) -> str:
        """
        Get the name of the S3 bucket from the URI.

        Returns
        -------
        str
            The name of the S3 bucket.
        """
        return self.config.bucket_name

    def _get_client(self) -> S3Client:
        """
        Get an S3 client object.

        Returns
        -------
        S3Client
            Returns a client object that interacts with the S3 storage service.
        """
        cfg = {
            "endpoint_url": self.config.endpoint_url,
            "aws_access_key_id": self.config.aws_access_key_id,
            "aws_secret_access_key": self.config.aws_secret_access_key,
        }
        return boto3.client("s3", **cfg)

    def _check_factory(self) -> tuple[S3Client, str]:
        """
        Check if the S3 bucket is accessible by sending a head_bucket request.

        Returns
        -------
        tuple[S3Client, str]
            A tuple containing the S3 client object and the name of the S3 bucket.
        """
        client = self._get_client()
        bucket = self._get_bucket()
        self._check_access_to_storage(client, bucket)
        return client, bucket

    @staticmethod
    def _check_access_to_storage(client: S3Client, bucket: str) -> None:
        """
        Check if the S3 bucket is accessible by sending a head_bucket request.

        Parameters
        ----------
        client : S3Client
            The S3 client object.
        bucket : str
            The name of the S3 bucket.

        Returns
        -------
        None

        Raises
        ------
        StoreError:
            If access to the specified bucket is not available.
        """
        try:
            client.head_bucket(Bucket=bucket)
        except ClientError as exc:
            raise StoreError("No access to s3 bucket!") from exc

    ############################
    # Private I/O methods
    ############################

    def _download_file(self, key: str, dst: str) -> Path:
        """
        Download a file from S3 based storage. The function checks if the bucket is accessible
        and if the destination directory exists. If the destination directory does not exist,
        it will be created.

        Parameters
        ----------
        key : str
            The key of the file on S3 based storage.
        dst : str
            The destination of the file on local filesystem.

        Returns
        -------
        Path
            The path of the downloaded file.
        """
        client, bucket = self._check_factory()
        client.download_file(bucket, key, dst)
        return Path(dst)

    def _get_presinged_url(self, src: str) -> str:
        """
        Returns a presigned URL for a specified object in an S3 bucket.

        Parameters:
        -----------
        src : str
            The name of the file.

        Returns:
        --------
        url: string
            A string representing the generated Presigned URL
        """
        client, bucket = self._check_factory()
        return client.generate_presigned_url(
            ClientMethod="get_object",
            Params={"Bucket": bucket, "Key": src},
            ExpiresIn=7200,
        )
